"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || (function () {
    var ownKeys = function(o) {
        ownKeys = Object.getOwnPropertyNames || function (o) {
            var ar = [];
            for (var k in o) if (Object.prototype.hasOwnProperty.call(o, k)) ar[ar.length] = k;
            return ar;
        };
        return ownKeys(o);
    };
    return function (mod) {
        if (mod && mod.__esModule) return mod;
        var result = {};
        if (mod != null) for (var k = ownKeys(mod), i = 0; i < k.length; i++) if (k[i] !== "default") __createBinding(result, mod, k[i]);
        __setModuleDefault(result, mod);
        return result;
    };
})();
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TestGenerator = void 0;
const path_1 = __importDefault(require("path"));
const generator_1 = require("../generator");
const fs = __importStar(require("fs/promises"));
const metaschema_1 = require("../metaschema");
const utils_1 = require("../utils");
const unsupported_aggregations_1 = require("./unsupported-aggregations");
const bson = __importStar(require("bson"));
class TestGenerator extends generator_1.GeneratorBase {
    schemaBsonTypeToTS(type) {
        switch (type) {
            case 'Binary':
                return 'bson.Binary';
            case 'Boolean':
                return 'boolean';
            case 'Code':
                return 'bson.Code';
            case 'CodeWScope':
                return 'bson.Code';
            case 'Date':
                return 'Date';
            case 'Decimal128':
                return 'bson.Decimal128';
            case 'Double':
                return 'bson.Double | number';
            case 'Int32':
                return 'bson.Int32 | number';
            case 'Long':
            case 'Int64':
                return 'bson.Long';
            case 'MaxKey':
                return 'bson.MaxKey';
            case 'MinKey':
                return 'bson.MinKey';
            case 'Null':
                return 'null';
            case 'ObjectId':
                return 'bson.ObjectId';
            case 'BSONRegExp':
                return 'bson.BSONRegExp';
            case 'String':
                return 'string';
            case 'BSONSymbol':
                return 'bson.BSONSymbol';
            case 'Timestamp':
                return `bson.Timestamp`;
            case 'Undefined':
                return `undefined`;
            case 'Number':
                return 'number';
            default:
                throw new Error(`Unknown BSON type: ${type}`);
        }
    }
    simplifiedTypesToTS(types) {
        const mappedTypes = [];
        for (const type of types) {
            if (type.bsonType === 'Document' && 'fields' in type) {
                mappedTypes.push(this.simplifiedSchemaToTS(type.fields));
            }
            else if (type.bsonType === 'Array' && 'types' in type) {
                mappedTypes.push(`Array<${this.simplifiedTypesToTS(type.types)}>`);
            }
            else {
                mappedTypes.push(this.schemaBsonTypeToTS(type.bsonType));
            }
        }
        return mappedTypes.join(' | ');
    }
    simplifiedSchemaToTS(schema) {
        let result = '{\n';
        for (const [key, value] of Object.entries(schema)) {
            result += `'${key}': ${this.simplifiedTypesToTS(value.types)};\n`;
        }
        result += '}\n';
        return result;
    }
    stageToTS(stage) {
        switch (typeof stage) {
            case 'object': {
                if (stage === null) {
                    return 'null';
                }
                if (Array.isArray(stage)) {
                    return `[${stage.map((s) => this.stageToTS(s)).join(', ')}]`;
                }
                if (stage instanceof bson.Binary) {
                    return `bson.Binary.createFromBase64('${stage.toString('base64')}', ${stage.sub_type})`;
                }
                if (stage instanceof bson.ObjectId) {
                    return `bson.ObjectId.createFromHexString('${stage.toHexString()}')`;
                }
                if (stage instanceof bson.Long) {
                    return `new bson.Long('${stage.toString()}')`;
                }
                if (stage instanceof Date) {
                    return `new Date('${stage.toISOString()}')`;
                }
                if (stage instanceof bson.BSONRegExp) {
                    return `new bson.BSONRegExp('${stage.pattern}', '${stage.options}')`;
                }
                if ('$code' in stage && typeof stage.$code === 'string') {
                    return JSON.stringify((0, utils_1.removeNewlines)(stage.$code));
                }
                let result = '{';
                for (const [key, value] of Object.entries(stage)) {
                    result += `${JSON.stringify(key)}: ${this.stageToTS(value)},`;
                }
                result += '}';
                return result;
            }
            case 'undefined':
                return 'undefined';
            default:
                return JSON.stringify(stage);
        }
    }
    emitTestBody(category, operator, test) {
        if (!test.pipeline) {
            this.emit(`// TODO: No pipeline found for ${operator}.${test.name}\n`);
            return;
        }
        if (!test.schema || typeof test.schema === 'string') {
            this.emit(`// TODO: no schema found for ${operator}.${test.name}${test.schema ? `: ${test.schema}` : ''}\n`);
            return;
        }
        const collectionName = Object.keys(test.schema)[0];
        const schema = test.schema[collectionName];
        this.emit(`type ${collectionName} = ${this.simplifiedSchemaToTS(schema)}\n`);
        this.emit(`const aggregation: schema.Pipeline<${collectionName}> = [\n`);
        for (let i = 0; i < test.pipeline.length; i++) {
            const stage = test.pipeline[i];
            const unsupportedStage = unsupported_aggregations_1.unsupportedAggregations[category]?.[operator]?.[test.name];
            const isUnsupportedStage = unsupportedStage && i >= unsupportedStage.stage;
            if (isUnsupportedStage) {
                this.emitComment(`This stage is unsupported by the static type system, so we're casting it to 'any' (${unsupportedStage.comment ?? 'it may involve a projected field'}).`);
            }
            this.emit(this.stageToTS(stage));
            if (isUnsupportedStage) {
                this.emit(' as any');
            }
            this.emit(',\n');
        }
        this.emit('];\n');
    }
    async generateImpl(yamlFiles) {
        for await (const file of yamlFiles) {
            const namespace = `${(0, utils_1.capitalize)(file.category)}Operators`;
            const basePath = path_1.default.resolve(__dirname, '..', '..', 'tests', file.category);
            await fs.mkdir(basePath, { recursive: true });
            for await (const operator of file.operators()) {
                const parsed = metaschema_1.Operator.parse(operator.yaml);
                const operatorName = parsed.name.replace(/^\$/, '');
                const filePath = path_1.default.join(basePath, `${operatorName}.spec.ts`);
                await this.emitToFile(filePath);
                this.emit('/* eslint-disable @typescript-eslint/no-unused-vars */\n');
                this.emit('/* eslint-disable filename-rules/match */\n');
                this.emit('/* eslint-disable @typescript-eslint/consistent-type-imports */\n');
                this.emit("import type * as schema from '../../out/schema';\n");
                this.emit("import * as bson from 'bson';\n\n");
                let i = 0;
                for (const test of parsed.tests ?? []) {
                    this.emitComment(test.name ?? `Test ${namespace}.${parsed.name}`, test.link);
                    this.emit(`function test${i++}() {\n`);
                    this.emitTestBody(file.category, operatorName, test);
                    this.emit('}\n\n');
                }
            }
        }
    }
}
exports.TestGenerator = TestGenerator;
//# sourceMappingURL=test-generator.js.map